.module/ram/abs=0         lpc2k4_through;

{  LPC2k4.DSP - talk through, encoding and decoding using LPC.

   Input: Speech samples from codec (using autobuffering) via sport0

   Ouput: Speech samples to speaker (using autobuffering) via sport0

   Modules used:
     - pre_emphasize_speech       (PREEMP.DSP)
     - gain_calculation           (GAIN.DSP)
     - autocorrelation_of_speech  (AUTOCOR.DSP)
     - durbin_double/single       (DURBIN2.DSP/DURBIN.DSP)
     - pitch_detection            (PITCH.DSP)
     - lpc_sync_synth             (SSYNTH.DSP)
     - de_emphasize_speech        (DEEMP.DSP)

     - constant header            (LPC.H)


   Description:
     This program implements a shell to demonstrate the LPC algorithm
     on an EZ-LAB board. Speech is autobuffered in from the codec,
     compressed, decompressed and autobuffered back out to the codec,
     providing a "talk-through" program.

}

.include    <..\system.k>;


{******************************************************************************
 *
 *  Variables
 *
 ******************************************************************************}

{include constant definitions}
.include <lpc.h>;

{Buffers used by autobuffering, input swaps between analys_buf & receive_buf,
 whereas output swaps between synth_buf & transit_buf}
.var/dm/ram/circ  analys_buf[FRAME_LENGTH];
.var/dm/ram/circ  receive_buf[FRAME_LENGTH];

{The LPC-parameters are stored in trans_line to "simulate" transmission}
.var/dm/ram       trans_line[WORDS_PR_LPCFRAME];

{Pointers to the buffers that are NOT currently being used by autobuffering}
.var/dm/ram       p2_analysis,p2_synth;

{Intermediate variables}
.var/pm/ram       autocor_speech[FRAME_LENGTH];
.var/dm/ram/circ  k[N];
.var/dm/ram       pitch, gain;
.var/dm/ram       lpc_flag;
.var/dm/ram       bufIx, bufLx, bufIs, bufLs;

.external   pre_emph;
.external   calc_gain;
.external   a_correlate;
.external   levinson;
.external   detect_pitch;
.external   encode;
.external   decode;
.external   clear_filter;
.external   synthesis;
.external   de_emph;

.VAR/DM     StateFlag, resetFlag;

.var/dm/ram/circ                rx_buf[3];      /* Status + L data + R data */
.var/dm/ram/circ                tx_buf[3];      /* Cmd + L data + R data    */
.var/dm/ram/circ                init_cmds[13];
.var/dm                         stat_flag;

.init resetFlag: 0;
.init tx_buf:   0xc000, 0x0000, 0x0000; /* Initially set MCE        */

.init init_cmds:
        0xc002,     {
                        Left input control reg
                        b7-6: 0=left line 1
                              1=left aux 1
                              2=left line 2
                              3=left line 1 post-mixed loopback
                        b5-4: res
                        b3-0: left input gain x 1.5 dB
                    }
        0xc102,     {
                        Right input control reg
                        b7-6: 0=right line 1
                              1=right aux 1
                              2=right line 2
                              3=right line 1 post-mixed loopback
                        b5-4: res
                        b3-0: right input gain x 1.5 dB
                    }
        0xc288,     {
                        left aux 1 control reg
                        b7  : 1=left aux 1 mute
                        b6-5: res
                        b4-0: gain/atten x 1.5, 08= 0dB, 00= 12dB
                    }
        0xc388,     {
                        right aux 1 control reg
                        b7  : 1=right aux 1 mute
                        b6-5: res
                        b4-0: gain/atten x 1.5, 08= 0dB, 00= 12dB
                    }
        0xc488,     {
                        left aux 2 control reg
                        b7  : 1=left aux 2 mute
                        b6-5: res
                        b4-0: gain/atten x 1.5, 08= 0dB, 00= 12dB
                    }
        0xc588,     {
                        right aux 2 control reg
                        b7  : 1=right aux 2 mute
                        b6-5: res
                        b4-0: gain/atten x 1.5, 08= 0dB, 00= 12dB
                    }
        0xc680,     {
                        left DAC control reg
                        b7  : 1=left DAC mute
                        b6  : res
                        b5-0: attenuation x 1.5 dB
                    }
        0xc780,     {
                        right DAC control reg
                        b7  : 1=right DAC mute
                        b6  : res
                        b5-0: attenuation x 1.5 dB
                    }
        0xc850,     {
                        data format register
                        b7  : res
                        b5-6: 0=8-bit unsigned linear PCM
                              1=8-bit u-law companded
                              2=16-bit signed linear PCM
                              3=8-bit A-law companded
                        b4  : 0=mono, 1=stereo
                        b0-3: 0=  8.
                              1=  5.5125
                              2= 16.
                              3= 11.025
                              4= 27.42857
                              5= 18.9
                              6= 32.
                              7= 22.05
                              8=   .
                              9= 37.8
                              a=   .
                              b= 44.1
                              c= 48.
                              d= 33.075
                              e=  9.6
                              f=  6.615
                       (b0) : 0=XTAL1 24.576 MHz; 1=XTAL2 16.9344 MHz
                    }
        0xc909,     {
                        interface configuration reg
                        b7-4: res
                        b3  : 1=autocalibrate
                        b2-1: res
                        b0  : 1=playback enabled
                    }
        0xca00,     {
                        pin control reg
                        b7  : logic state of pin XCTL1
                        b6  : logic state of pin XCTL0
                        b5  : master - 1=tri-state CLKOUT
                              slave  - x=tri-state CLKOUT
                        b4-0: res
                    }
        0xcc40,     {
                        miscellaneous information reg
                        b7  : 1=16 slots per frame, 0=32 slots per frame
                        b6  : 1=2-wire system, 0=1-wire system
                        b5-0: res
                    }
        0xcd00;     {
                        digital mix control reg
                        b7-2: attenuation x 1.5 dB
                        b1  : res
                        b0  : 1=digital mix enabled
                    }

{******************************************************************************
 *
 *  Interrupt vector table
 *
 ******************************************************************************}
        jump start;  rti; rti; rti;     {00: reset }
        rti;         rti; rti; rti;     {04: IRQ2 }
        rti;         rti; rti; rti;     {08: IRQL1 }
        rti;         rti; rti; rti;     {0c: IRQL0 }
        ar = dm(stat_flag);             {10: SPORT0 tx }
        ar = pass ar;
        if eq rti;
        jump next_cmd;
        jump input_samples;             {14: SPORT1 rx }
                     rti; rti; rti;
        jump irqeisr;
                     rti; rti; rti;     {18: IRQE }
        rti;         rti; rti; rti;     {1c: BDMA }
        jump irq1isr;
                     rti; rti; rti;     {20: SPORT1 tx or IRQ1 }
        rti;         rti; rti; rti;     {24: SPORT1 rx or IRQ0 }
        rti;         rti; rti; rti;     {28: timer }
        rti;         rti; rti; rti;     {2c: power down }


{******************************************************************************
 *
 *  ADSP 2181 intialization
 *
 ******************************************************************************}
start:
        {   shut down sport 0 }
        ax0 = b#0000100000000000;   dm (System_Control_Reg) = ax0;

        dis timer;

  ax0 = 0x0000;
  DM(StateFlag) = AX0;

        i3 = ^rx_buf;
        l3 = %rx_buf;
        i7 = ^tx_buf;
        l7 = %tx_buf;
        i0 = ^init_cmds;
        l0 = %init_cmds;

        m1 = 1;
        m5 = 1;


{================== S E R I A L   P O R T   #0   S T U F F ==================}
        ax0 = b#0000111010110111;   dm (SPORT0_Autobuf) = ax0;
            {   |||!|-/!/|-/|/|+- receive autobuffering 0=off, 1=on
                |||!|  ! |  | +-- transmit autobuffering 0=off, 1=on
                |||!|  ! |  +---- | receive m?
                |||!|  ! |        | m1
                |||!|  ! +------- ! receive i?
                |||!|  !          ! i3
                |||!|  !          !
                |||!|  +========= | transmit m?
                |||!|             | m5
                |||!+------------ ! transmit i?
                |||!              ! i7
                |||!              !
                |||+============= | BIASRND MAC biased rounding control bit
                ||+-------------- 0
                |+--------------- | CLKODIS CLKOUT disable control bit
                +---------------- 0
            }

        ax0 = 0;    dm (SPORT0_RFSDIV) = ax0;
            {   RFSDIV = SCLK Hz/RFS Hz - 1 }
        ax0 = 0;    dm (SPORT0_SCLKDIV) = ax0;
            {   SCLK = CLKOUT / (2  (SCLKDIV + 1) }
        ax0 = b#1000011000001111;   dm (SPORT0_Control_Reg) = ax0;
            {   multichannel
                ||+--/|!||+/+---/ | number of bit per word - 1
                |||   |!|||       | = 15
                |||   |!|||       |
                |||   |!|||       |
                |||   |!||+====== ! 0=right just, 0-fill; 1=right just, signed
                |||   |!||        ! 2=compand u-law; 3=compand A-law
                |||   |!|+------- receive framing logic 0=pos, 1=neg
                |||   |!+-------- transmit data valid logic 0=pos, 1=neg
                |||   |+========= RFS 0=ext, 1=int
                |||   +---------- multichannel length 0=24, 1=32 words
                ||+-------------- | frame sync to occur this number of clock
                ||                | cycle before first bit
                ||                |
                ||                |
                |+--------------- ISCLK 0=ext, 1=int
                +---------------- multichannel 0=disable, 1=enable
            }
            {   non-multichannel
                |||!|||!|||!+---/ | number of bit per word - 1
                |||!|||!|||!      | = 15
                |||!|||!|||!      |
                |||!|||!|||!      |
                |||!|||!|||+===== ! 0=right just, 0-fill; 1=right just, signed
                |||!|||!||+------ ! 2=compand u-law; 3=compand A-law
                |||!|||!|+------- receive framing logic 0=pos, 1=neg
                |||!|||!+-------- transmit framing logic 0=pos, 1=neg
                |||!|||+========= RFS 0=ext, 1=int
                |||!||+---------- TFS 0=ext, 1=int
                |||!|+----------- TFS width 0=FS before data, 1=FS in sync
                |||!+------------ TFS 0=no, 1=required
                |||+============= RFS width 0=FS before data, 1=FS in sync
                ||+-------------- RFS 0=no, 1=required
                |+--------------- ISCLK 0=ext, 1=int
                +---------------- multichannel 0=disable, 1=enable
            }


        ax0 = b#0000000000000111;   dm (SPORT0_TX_Channels0) = ax0;
            {   ^15          00^   transmit word enables: channel # == bit # }
        ax0 = b#0000000000000111;   dm (SPORT0_TX_Channels1) = ax0;
            {   ^31          16^   transmit word enables: channel # == bit # }
        ax0 = b#0000000000000111;   dm (SPORT0_RX_Channels0) = ax0;
            {   ^15          00^   receive word enables: channel # == bit # }
        ax0 = b#0000000000000111;   dm (SPORT0_RX_Channels1) = ax0;
            {   ^31          16^   receive word enables: channel # == bit # }


{============== S Y S T E M   A N D   M E M O R Y   S T U F F ==============}
        ax0 = b#0001100000000000;   dm (System_Control_Reg) = ax0;
            {   +-/!||+-----/+-/- | program memory wait states
                |  !|||           | 0
                |  !|||           |
                |  !||+---------- 0
                |  !||            0
                |  !||            0
                |  !||            0
                |  !||            0
                |  !||            0
                |  !||            0
                |  !|+----------- SPORT1 1=serial port, 0=FI, FO, IRQ0, IRQ1,..
                |  !+------------ SPORT1 1=enabled, 0=disabled
                |  +============= SPORT0 1=enabled, 0=disabled
                +---------------- 0
                                  0
                                  0
            }



        ifc = b#00000011111111;         { clear pending interrupt }
        nop;


        icntl = b#00010;
            {     ||||+- | IRQ0: 0=level, 1=edge
                  |||+-- | IRQ1: 0=level, 1=edge
                  ||+--- | IRQ2: 0=level, 1=edge
                  |+---- 0
                  |----- | IRQ nesting: 0=disabled, 1=enabled
            }


        mstat = b#1000000;
            {     ||||||+- | Data register bank select
                  |||||+-- | FFT bit reverse mode (DAG1)
                  ||||+--- | ALU overflow latch mode, 1=sticky
                  |||+---- | AR saturation mode, 1=saturate, 0=wrap
                  ||+----- | MAC result, 0=fractional, 1=integer
                  |+------ | timer enable
                  +------- | GO MODE
            }





{******************************************************************************
 *
 *  ADSP 1847 Codec intialization
 *
 ******************************************************************************}

        {   clear flag }
        ax0 = 1;
        dm(stat_flag) = ax0;

        {   enable transmit interrupt }
        ena ints;
        imask = b#0001000000;
            {     |||||||||+ | timer
                  ||||||||+- | SPORT1 rec or IRQ0
                  |||||||+-- | SPORT1 trx or IRQ1
                  ||||||+--- | BDMA
                  |||||+---- | IRQE
                  ||||+----- | SPORT0 rec
                  |||+------ | SPORT0 trx
                  ||+------- | IRQL0
                  |+-------- | IRQL1
                  +--------- | IRQ2
            }


        ax0 = dm (i7, m5);          { start interrupt }
        tx0 = ax0;

check_init:
        ax0 = dm (stat_flag);       { wait for entire init }
        af = pass ax0;              { buffer to be sent to }
        if ne jump check_init;      { the codec            }

        ay0 = 2;
check_aci1:
        ax0 = dm (rx_buf);          { once initialized, wait for codec }
        ar = ax0 and ay0;           { to come out of autocalibration }
        if eq jump check_aci1;      { wait for bit set }

check_aci2:
        ax0 = dm (rx_buf);          { wait for bit clear }
        ar = ax0 and ay0;
        if ne jump check_aci2;
        idle;

        ay0 = 0xbf3f;               { unmute left DAC }
        ax0 = dm (init_cmds + 6);
        ar = ax0 AND ay0;
        dm (tx_buf) = ar;
        idle;

        ax0 = dm (init_cmds + 7);   { unmute right DAC }
        ar = ax0 AND ay0;
        dm (tx_buf) = ar;
        idle;


        ifc = b#00000011111111;     { clear any pending interrupt }
        nop;


        imask = b#0000110100;       { enable rx0 interrupt }
            {     |||||||||+ | timer
                  ||||||||+- | SPORT1 rec or IRQ0
                  |||||||+-- | SPORT1 trx or IRQ1
                  ||||||+--- | BDMA
                  |||||+---- | IRQE
                  ||||+----- | SPORT0 rec
                  |||+------ | SPORT0 trx
                  ||+------- | IRQL0
                  |+-------- | IRQL1
                  +--------- | IRQ2
            }


{------------------------------------------------------------------------------
 -
 -  command loop.
 -
 ------------------------------------------------------------------------------}
start_test:
 {Default register values, these values can always be asumed, and must
  be reset if altered}
  l0 = 0; l1 = 0; l2 = 0;
  l4 = 0; l5 = 0; l6 = 0;

 {DEDICATED REGISTERS. These registers must NOT be altered by any
  routine at any time! (used by autobuffering)}
  m0 = 0;  m4 = 0;
  m1 = 1;  m5 = 1;
{  i3 = ^receive_buf;    l3 = %receive_buf;}
  ax0 = ^receive_buf;   dm (bufIx) = ax0;
  ax0 = %receive_buf;   dm (bufLx) = ax0;

 {Setup and clear intermediate buffers}
  i6 = ^analys_buf;     l6 = 0;
  dm(p2_analysis) = i6;

  ena sec_reg;
  ax0 = FRAME_LENGTH;
  af  = pass ax0;
  dis sec_reg;

  ax0 = 0;
  dm(lpc_flag) = ax0;

 {clear the synthesis filter}
  call clear_filter;

FirstState: IDLE;
 ax0 = dm(resetFlag);   { set this variable to non-zero to return to host }
 ar  = pass ax0;
 if ne rts;
            AR = DM(StateFlag);
            AR = PASS AR;
            IF EQ JUMP FirstState;

wait: idle;

 ax0 = dm(resetFlag);   { set this variable to non-zero to return to host }
 ar  = pass ax0;
 if ne rts;

 ax0 = dm(lpc_flag);
 ar  = pass ax0;
 if eq jump wait;

  ax0 = 0;
  dm(lpc_flag) = ax0;

  {Parameters: i0 = p2_analysis (-> speech)
  Returns   : filtered speech}
  i0 = dm(p2_analysis);  l0 = 0;
  call pre_emph;

 {Parameters: i0  = p2_analysis (-> speech)
  Returns   : sr1 = gain}
  i0 = dm(p2_analysis); l0 = 0;
  call calc_gain;
  dm(gain) = sr1;

 {Parameters: i0 = p2_analysis (-> speech)
  Returns   : autocor_speech[]}
  i0 = dm(p2_analysis); l0 = 0;
  i6 = ^autocor_speech; l6 = 0;
  call a_correlate;

 {Parameters: i4 -> autocor_speech[])
  Returns   : i0 -> k[]}
  i4 = ^autocor_speech; l4 = 0;
  i0 = ^k;              l0 = 0;
  call levinson;

  {Parameters: i0 -> k[],
               i6 -> autocor_speech[]
   Returns   : si =  pitch}
  i0 = ^k;              l0 = 0;
  i6 = ^autocor_speech; l6 = 0;
  call detect_pitch;
  dm(pitch) = si;

  {Parameters: i1 -> k[],
               ar =  pitch,
               si =  gain
   Returns   : parameters encoded
               ar =  pitch,
               si =  gain}
  i1 = ^k;   l1 = 0;
  ar = dm(pitch);
  si = dm(gain);
  call encode;
  dm(pitch) = ar;
  dm(gain)  = si;

{TRANSMISSION LINE}

 {Parameters: i1 -> k[],
              si  = pitch,
              ax0 = gain
   Returns  : k's decoded
              si  = pitch,
              ax0 = gain}
  i1  = ^k;   l1 = 0;
  si  = dm(pitch);
  ax0 = dm(gain);
  call decode;
  dm(pitch) = si;
  dm(gain)  = ax0;

 {Parameters: ax1 =  pitch,
              mx1 =  gain,
              i0  -> k[]
  Returns   : i2  -> speech}
  i0 = ^k;           l0 = 0;
  i1 = ^k + N - 1;   l1 = 0; {store k's in revers order - }
  cntr = 5; {N/2}            {required by lattice routine}
  m2 = -1;
  do reverse_ks until ce;
    ay0 = dm(i0,m0);
    ay1 = dm(i1,m0);
    dm(i0,m1) = ay1;
  reverse_ks: dm(i1,m2) = ay0;
  ax1 = dm(pitch);
  mx1 = dm(gain);
  i1 = ^k;           l1 = N;
  i2 = dm(p2_analysis); l2 = 0;
  call synthesis;

 {Parameters: i0 = p2_analysis (-> speech)
  Returns   : filtered speech}
  i0 = dm(p2_analysis);  l0 = 0;
  call de_emph;

jump wait;


{End of main routine}

{Autobuffering interrupt routines}


input_samples:
rcv_ir:

 ena sec_reg;
 AX0 = AR;
 AY0 = DM(StateFlag);
 AR = PASS AY0;
 IF EQ JUMP TalKThru;
 set fl1;                     { indicate LPC active }

 AR = AX0;
 { save i0 and l0 for input buffering }
 dm (bufIs) = i0;
 dm (bufLs) = l0;

 i0 = dm (bufIx);
 l0 = dm (bufLx);


 ar = dm(i0,m0);
 sr = ashift ar by 2 (hi);
 dm (tx_buf + 1) = sr1;
 dm (tx_buf + 2) = sr1;
 ar = dm (rx_buf + 2);
 sr = ashift ar by -2 (hi);
 dm(i0,m1) = sr1;

 dm (bufIx) = i0;

 { restore i0 and l0 after using for input buffering }
 i0 = dm (bufIs);
 l0 = dm (bufLs);

 af = af - 1;
 if gt jump no_lpc;
  {switch pointers}
  ay0 = dm (bufIx);
  ay1 = dm(p2_analysis);
  dm (bufIx) = ay1;
  dm(p2_analysis) = ay0;

  ax0 = FRAME_LENGTH;
  af  = pass ax0;

  ax0 = 1;
  dm(lpc_flag) = ax0;
 no_lpc:

 dis sec_reg;
 rti;

TalKThru:
   reset fl1;               { indicate LPC inactive }
        ax0 = dm (rx_buf + 2);      { loopback inputs to outputs }
        dm (tx_buf + 1) = ax0;
        dm (tx_buf + 2) = ax0;

          DIS Sec_Reg;
          RTI;

 {END of main code}

{------------------------------------------------------------------------------
 -
 -  transmit interrupt used for Codec initialization
 -
 ------------------------------------------------------------------------------}
next_cmd:
        ena sec_reg;
        ax0 = dm (i0, m1);          { fetch next control word and }
        dm (tx_buf) = ax0;          { place in transmit slot 0    }
        ax0 = i0;
        ay0 = ^init_cmds;
        ar = ax0 - ay0;
        if gt rti;                  { rti if more control words still waiting }
        ax0 = 0xaf00;               { else set done flag and }
        dm (tx_buf) = ax0;          { remove MCE if done initialization }
        ax0 = 0;
        dm (stat_flag) = ax0;       { reset status flag }
        rti;

irq1isr:
        AX0 = 1;
        dm(resetFlag) = AX0;   { set this variable to non-zero to return to host }
        RTI;

irqeisr:
        AX0 = ar;               { not sure, better preserve ar }
        ar = DM(StateFlag);
        ar = not ar;
        DM(StateFlag) = ar;
        ar = ax0;
        rti;


.endmod;


